/**
 * vim: set ts=4 :
 * =============================================================================
 * MegaHAL Extension
 * Copyright (C) 2011 Afronanny.  All rights reserved.
 * =============================================================================
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License, version 3.0, as published by the
 * Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "megahal.h"
#include "extension.h"
#include <malloc.h>
#include <amtl/am-string.h>

/**
 * @file extension.cpp
 * @brief Implement extension code here.
 */

MHExt g_MHExt;		/**< Global singleton for extension's main interface */

SMEXT_LINK(&g_MHExt);

cell_t DoReply(IPluginContext *pContext, const cell_t *params);
cell_t LearnWithoutReply(IPluginContext *pContext, const cell_t *params);

//This is a bad native - crashes servers often
//cell_t MH_Command(IPluginContext *pContext, const cell_t *params);

const sp_nativeinfo_t Natives[] = 
{
	{"MH_DoReply",		DoReply},
	{"MH_Learn",		LearnWithoutReply},
//	{"MH_Command",	MH_Command},
	{NULL,				NULL},
};

bool MHExt::SDK_OnLoad(char *error, size_t maxlength, bool late)
{
	char dir[PLATFORM_MAX_PATH];
	g_pSM->BuildPath(Path_SM, dir, sizeof(dir), "megahal-ext");
	megahal_setdirectory(dir);
	megahal_seterrorfile("error.log");
	megahal_setstatusfile("status.log");

	megahal_initialize();

	sharesys->AddNatives(myself, Natives);
	sharesys->RegisterLibrary(myself, "MegaHAL Ext");
	return true;
}

void MHExt::SDK_OnUnload()
{
	megahal_cleanup();
}

typedef struct threadparams {
	ke::AString phrase;
	ke::AString reply;
	IPluginFunction *pFunc;
} threadparams_t;

static void RespondToPlugin(void *pData) {
	threadparams_t *params = (threadparams_t*)pData;

	params->pFunc->PushString(params->reply.chars());
	params->pFunc->Execute(0);

	delete pData;
}

class MegahalThread : public IThread
{
public:
	MegahalThread(threadparams_t *param) :
		m_param(param)
	{
	}

	void RunThread(IThreadHandle *pHandle) {
		m_param->reply = megahal_do_reply((char *)m_param->phrase.chars(), 1);
		smutils->AddFrameAction(&RespondToPlugin, m_param);
	}

	void OnTerminate(IThreadHandle *pHandle, bool cancel) {
		if (cancel)
			delete m_param;
	}

private:
	threadparams_t *m_param;
};
	
cell_t DoReply(IPluginContext *pContext, const cell_t *params)
{
	char *phrase;

	pContext->LocalToString(params[1], &phrase);

	IPluginFunction *pFunc = pContext->GetFunctionById(params[2]);
	if (!pFunc) {
		return pContext->ThrowNativeError("Invalid function id %d", params[2]);
	}

	threadparams_t *param = new threadparams_t;
	param->phrase = phrase;
	param->reply = nullptr;
	param->pFunc = pFunc;

	//The thread is there because megahal isn't as optimized as it could be
	//and causes a 80ms hangup while it generates a reply
	threader->MakeThread(new MegahalThread(param));

	return 1;
}

//BAD BAD BAD - DO NOT USE THIS.
/*cell_t MH_Command(IPluginContext *pContext, const cell_t *params)
{
	char *input;
	pContext->LocalToString(params[1], &input);
	return megahal_command(input);
}*/

cell_t LearnWithoutReply(IPluginContext *pContext, const cell_t *params)
{
	char *input;

	pContext->LocalToString(params[1], &input);

	megahal_learn_no_reply(input, 1);
	return 1;
}